program benchio

  use benchclock
  use mpiio

  implicit none

! Set local array size - global sizes l1, l2 and l3 are scaled
! by number of processes in each dimension

  integer, parameter :: n1 = 128
  integer, parameter :: n2 = 128
  integer, parameter :: n3 = 128

  integer :: i1, i2, i3, j1, j2, j3, l1, l2, l3, p1, p2, p3

  double precision :: iodata(0:n1+1, 0:n2+1, 0:n3+1)

  integer :: rank, size, ierr, comm, cartcomm, dblesize
  integer, dimension(ndim) :: dims, coords

  integer, parameter :: iounit = 12
  integer, parameter :: mib = 1024*1024

  logical :: reorder = .false.
  logical, dimension(ndim) :: periods = [.false., .false., .false.]

  double precision :: t0, t1, time, iorate, mibdata

  call MPI_Init(ierr)

  comm = MPI_COMM_WORLD

  call MPI_Comm_size(comm, size, ierr)
  call MPI_Comm_rank(comm, rank, ierr)

  dims(:) = 0

! Set 3D processor grid

  call MPI_Dims_create(size, ndim, dims, ierr)

! Reverse dimensions as MPI assumes C ordering (this is not essential)

  p1 = dims(3)
  p2 = dims(2)
  p3 = dims(1)

! Compute global sizes

  l1 = p1*n1
  l2 = p2*n2
  l3 = p3*n3

  call MPI_Type_size(MPI_DOUBLE_PRECISION, dblesize, ierr)

  mibdata = float(dblesize*n1*n2*n3)*float(p1*p2*p3)/float(mib)

  if (rank == 0) then
     write(*,*)
     write(*,*) 'Simple MPI-IO benchmark'
     write(*,*) '-----------------------'
     write(*,*)
     write(*,*) 'Running on ', size, ' process(es)'
     write(*,*) 'Process grid is (', p1, ', ', p2, ', ', p3, ')'
     write(*,*) 'Array size is   (', n1, ', ', n2, ', ', n3, ')'
     write(*,*) 'Global size is  (', l1, ', ', l2, ', ', l3, ')'
     write(*,*)
     write(*,*) 'Total amount of data = ', mibdata, ' MiB'
     write(*,*)
     write(*,*) 'Clock resolution is ', benchtick()*1.0e6, ', usecs'
  end if
  
  dims(1) = p1
  dims(2) = p2
  dims(3) = p3

  call MPI_Cart_create(comm, ndim, dims, periods, reorder, cartcomm, ierr)


! Set halos to illegal values

  iodata(:,:,:) = -1
  
! Set iodata core to have unique values 1, 2, ..., p1*n1*p2*n2*p3*n3

  call MPI_Cart_coords(cartcomm, rank, ndim, coords, ierr)
  
  do i3 = 1, n3
     do i2 = 1, n2
        do i1 = 1, n1

           j1 = coords(1)*n1 + i1
           j2 = coords(2)*n2 + i2
           j3 = coords(3)*n3 + i3

           iodata(i1,i2,i3) = (j3-1)*l1*l2 + (j2-1)*l1 + j1

        end do
     end do
  end do

  if (rank == 0) then
     write(*,*)
     write(*,*) 'Parallel IO'
     write(*,*) '-----------'
  end if

  call MPI_Barrier(comm, ierr)
  t0 = benchtime()
  call mpiiowrite('unstriped/mpiionative.dat', iodata, n1, n2, n3, cartcomm)
  call MPI_Barrier(comm, ierr)
  t1 = benchtime()

  time = t1 - t0
  iorate = mibdata/time

  if (rank == 0) then
     write(*,*) 'Unstriped time = ', time, ', IO rate = ', iorate, ' MiB/s'
     call fdelete('unstriped/mpiionative.dat')
  end if

  call MPI_Barrier(comm, ierr)
  t0 = benchtime()
  call mpiiowrite('striped/mpiionative.dat', iodata, n1, n2, n3, cartcomm)
  call MPI_Barrier(comm, ierr)
  t1 = benchtime()
  
  time = t1 - t0
  iorate = mibdata/time

  if (rank == 0) then
     write(*,*) 'Striped time =   ', time, ', IO rate = ', iorate, ' MiB/s'
     call fdelete('striped/mpiionative.dat')
  end if

  call MPI_Barrier(comm, ierr)
  t0 = benchtime()
  call mpiiowrite('defstriped/mpiionative.dat', iodata, n1, n2, n3, cartcomm)
  call MPI_Barrier(comm, ierr)
  t1 = benchtime()
  
  time = t1 - t0
  iorate = mibdata/time

  if (rank == 0) then
     write(*,*) 'Defstripe time = ', time, ', IO rate = ', iorate, ' MiB/s'
     call fdelete('defstriped/mpiionative.dat')       
  end if


  if (rank == 0) then
     write(*,*)
     write(*,*) 'Serial IO'
     write(*,*) '---------'
  end if

  call MPI_Barrier(comm, ierr)
  t0 = benchtime()
  call serialwrite('unstriped/serial.dat', iodata, n1, n2, n3, cartcomm)
  call MPI_Barrier(comm, ierr)
  t1 = benchtime()

  time = t1 - t0
  iorate = mibdata/time

  if (rank == 0) then
     write(*,*) 'Unstriped time = ', time, ', IO rate = ', iorate, ' MiB/s'
     call fdelete('unstriped/serial.dat')
  end if

  call MPI_Barrier(comm, ierr)
  t0 = benchtime()
  call serialwrite('striped/serial.dat', iodata, n1, n2, n3, cartcomm)
  call MPI_Barrier(comm, ierr)
  t1 = benchtime()
  
  time = t1 - t0
  iorate = mibdata/time

  if (rank == 0) then
     write(*,*) 'Striped time =   ', time, ', IO rate = ', iorate, ' MiB/s'
     call fdelete('striped/serial.dat')
  end if

  call MPI_Barrier(comm, ierr)
  t0 = benchtime()
  call serialwrite('defstriped/serial.dat', iodata, n1, n2, n3, cartcomm)
  call MPI_Barrier(comm, ierr)
  t1 = benchtime()
  
  time = t1 - t0
  iorate = mibdata/time

  if (rank == 0) then
     write(*,*) 'Defstripe time = ', time, ', IO rate = ', iorate, ' MiB/s'
     call fdelete('defstriped/serial.dat')
  end if


  if (rank == 0) then
     write(*,*)
     write(*,*) 'Finished'
     write(*,*) '--------'
     write(*,*)
  end if

  call MPI_Finalize(ierr)
  
end program benchio

subroutine fdelete(filename)

  implicit none

  character *(*) :: filename
  integer, parameter :: iounit = 15
  integer :: stat

  write(*,*) 'Deleting: ', filename

  open(unit=iounit, iostat=stat, file=filename, status='old')
  if (stat.eq.0) close(unit=iounit, status='delete')

end subroutine fdelete
